#!/usr/bin/python3

import sys, os
import difflib
import webbrowser
from tkinter import Tk
from tkinter.filedialog import askdirectory

print("""
Das Programm sammelt alle Textdateien im Abgabeordner, vergleicht sie
untereinander und bei Bedarf auch mit allen Dateien in einem oder mehreren
nacheinander zu wählenden Archivordnern und sortiert sie nach Ähnlichkeit.
""")

# Dipl.-Ing. Martin Vogel
# Hochschule Bochum
# martin.vogel@hs-bochum.de
# 2016-11-16
# Lizenz: CC-by-3.0 https://creativecommons.org/licenses/by/3.0/de/

# Auswahl des zu untersuchenden Dateityps. Es ist wenig sinnvoll, mehrere
# Typen miteinander zu vergleichen, da die Inhalte normalerweise zu
# unterschiedlich sind.
# PDF- und Office-Dateien müssen vor dem Vergleich erst ins Textformat
# umgewandelt werden, unter Linux z.B. innerhalb eines Verzeichnisses mit
# „for f in *.pdf; do pdftotext "$f"; done“ oder innerhalb eines ganzen
# Verzeichnisbaums mit
# „for f in $(find . -iname *.pdf); do echo "$f"; pdftotext "$f"; done“
Dateityp = ".py"
print("Es wird nach Dateien gesucht, deren Name auf",Dateityp,"endet.")

# Das Programm hat (noch) keine grafische Oberfläche.
# Tkinter wird nur für die Verzeichnisauswahldialoge verwendet.
Fenster = Tk()
# Deshalb nehmen wir das GUI-Fenster wieder vom Schirm:
Fenster.withdraw() 

# Maximal auszuwertende Datenmenge pro Datei in Byte festlegen.
# Die Angabe kann sinnvoll sein, wenn das Programm zu lange braucht, um
# Ergebnisse zu liefern.
# In größeren Dateien wird dann nur der Anfang untersucht.
# -1 heißt: Datei immer komplett lesen.
maxsize = -1
if maxsize <1:
    print("Alle Dateien werden vollständig untersucht.")
else:
    print("Untersucht werden die ersten",maxsize,"Byte jeder Datei.")
    
# Soll schnell oder lieber gründlich untersucht werden?
# Wahlweise SequenceMatcher.ratio oder SequenceMatcher.quick_ratio verwenden
quick = False
print("Es erfolgt eine","schnelle" if quick else "gründliche",
      "Suche nach Übereinstimmungen.")

# Maximal auszugebende Zahl von Übereinstimmungen
maxresults = 50
print("Die",maxresults,"ähnlichsten Dateipaare werden angezeigt.")

# Wo sind die neu hinzugekommenen Dateien?    
Abgabeordner = askdirectory(
    initialdir=".",
    title='Bitte den Abgabeordner der neuen Dateien wählen:')
if not Abgabeordner:
    print("Kein Abgabeordner angegeben. Abbruch.")
    sys.exit()
    
# Dictionary aus Dateinamen und Dateiinhalten
data={}

# Vergleichsliste: [Ähnlichkeit, Dateiname1, Dateiname2]
comptab=[]

# Whitespace zusammenfassen
def unwhite(T):
    # Leerzeichen am Zeilenende entfernen
    while " \n" in T:
        T = T.replace(" \n","\n")
    # Leerzeilen entfernen
    while "\n\n" in T:
        T = T.replace("\n\n","\n")
    return T

# Pfadnamen für Tabellenausgabe kürzen
def kurzpfad(P,n=79):
    # Muss der überhaupt gekürzt werden?
    if len(P)<=n:
        return P
    # Der Dateiname selbst sollte möglichst erhalten bleiben.
    # Hier ist er:
    N = P.split("/")[-1]
    # Wenn der jedoch schon zu lang ist, kommt das Messer jetzt doch …
    if len(N)>=n-1:
        return "…"+N[-n+1:]
    # Standard: Pfadanfang…/Dateiname
    return P[:n-len(N)-2]+"…/"+N

# Wir nehmen mal an, dass die Ordner üblicherweise nicht allzuweit
# voneinander entfernt sind …
Archivordner = Abgabeordner

# Archivordner müssen nicht angegeben werden, aber solange welche
# gewählt werden, erfolgt eine weitere Frage nach einzulesenden Ordnern.
while Archivordner:
    Archivordner = askdirectory(
        initialdir=Archivordner+"/..", 
        title='Bitte (weiteren) Archivordner mit alten Dateien wählen:')
    if Archivordner:
        print("\nLese Archivordner:\n",Archivordner," …\n",sep="")
        for root,dir,files in os.walk(Archivordner):
            for f in files:
                if f.lower().endswith(Dateityp):
                    rf = os.path.join(root,f)
                    # Ein Punkt pro Datei …
                    print(".",end="")
                    with open(rf) as file:
                        data[rf] = unwhite(file.read(maxsize))
            print()

# Jetzt nacheinander Datei für Datei im Abgabeordner untersuchen:
print("\nDurchsuche Abgabeordner:\n",Abgabeordner," …\n",sep="")
S = difflib.SequenceMatcher()
for root,dir,files in os.walk(Abgabeordner):
    for f in files:
        if f.lower().endswith(Dateityp):
            rf = os.path.join(root,f)
            print(kurzpfad(rf,79))
            with open(rf) as file:
                data[rf] = unwhite(file.read(maxsize))
            # Zuletzt gelesene Datei mit allen vorher gelesenen
            # vergleichen:
            # "SequenceMatcher computes and caches detailed
            # information about the second sequence, so if you
            # want to compare one sequence against many sequences,
            # use set_seq2() to set the commonly used sequence once
            # and call set_seq1() repeatedly, once for each of the
            # other sequences." 
            S.set_seq2(data[rf])
            for g in data:
                # Datei nicht mit sich selbst vergleichen :-)
                if rf != g:
                    S.set_seq1(data[g])
                    # Bestimmung des Ähnlichkeitswerts:
                    # Genauere Werte: S.ratio()
                    # Schnellere Ergebnisse: S.quick_ratio()
                    if quick:
                        comptab.append([S.quick_ratio(),rf,g])
                    else:
                        comptab.append([S.ratio(),rf,g])

print("\nEinlesen und Analyse abgeschlossen.\n")
print("-"*79)
print("\nÄhnlichkeitstabelle:\n")

with open("Ähnlichkeit.html","w") as html:
    print("""<html><head><meta charset="utf-8"/><style>
table {font-family:monospace;
       border-color:lightgray;}
td.diff_header {color:gray;}
th.diff_header {color:darkred;}
colgroup {border-color:lightgray;}
h1 {color:darkred;}
</style></head><body>
<h1>Ähnlichkeitsanalyse</h1>
""",file=html)
    n = 1
    for s in sorted(comptab,reverse=True):
        print("%3i%% %37s %37s"%(s[0]*100,kurzpfad(s[1],37),kurzpfad(s[2],37)))
        print("<h2>%i Prozent Übereinstimmung</h2>"%(s[0]*100),file=html)
        print(difflib.HtmlDiff().make_table(data[s[1]].split("\n"),
                                            data[s[2]].split("\n"),
                                            s[1],s[2]),file=html)
        # Die Tabelle soll nicht uferlos wachsen
        n += 1
        if n > maxresults:
            break
    print("</body></html>",file=html)

webbrowser.open("Ähnlichkeit.html")    
